--File: AdderTree_rtl.vhd (vhdl)
--Complex tree adder of required size
--(c) Kongsberg Defence & Aerospace 2015
--$Id: ComplexAdderTree_rtl.vhd 14777 2015-03-24 10:58:36Z mathiab $
--Author: Mathias By
--
--Assumptions: 
-- <Add Assumptions>
--Description: 
-- The ComplexAdderTree module takes a array of N 16 bit Addends. A adder tree of the required
-- is used to make a as fast as possible and fully pipelined addition of the addends. The 20 bit result is not truncated,
--so if N > 16 the outputwidt have to be adjusted to make sure that overflow is avoided. 


architecture rtl of ComplexAdderTree is

   ----------------------------------------------------------------------------
   -- Types, constants and signals
   ----------------------------------------------------------------------------
   signal CalcStages : t_MxN20(0 to logCeil(2,c_NumAddends), 0 to c_NumAddends-1);--Holds all temporary results
   signal DataValidShift : std_logic_vector (logCeil(2,c_NumAddends) downto 0); --used to make ResultValid on the
                                                                                    --output reflect DataValid on the
                                                                                    --input after correct ammount of
                                                                                    --Clk cycles
   ----------------------------------------------------------------------------
   -- Procedures and functions
   ----------------------------------------------------------------------------

begin

   ----------------------------------------------------------------------------
   -- Combinatorial processes
   ----------------------------------------------------------------------------
   
   
   Result <= CalcStages(logCeil(2,c_NumAddends),0); 
   ResultValid <= DataValidShift(0);

   ----------------------------------------------------------------------------
   -- Clocked processes
   ----------------------------------------------------------------------------
   
   p_MyProcess : process(Clk) is
      --number of addends left
      variable v_n : integer;
      --number of addends in next calculation stage
      variable v_nNext :integer;
   
   begin
      if rising_edge(Clk) then
         if Rst_n = '0' then
            DataValidShift <= (others => '0') ;
            CalcStages <= (others =>(others =>(others =>(others => '0'))));  
         else
            --Handling data valid for the ResultValid input to reflect the DataValid input
            DataValidShift(DataValidShift'length-1) <= DataValid;
            DataValidShift(DataValidShift'length-2 downto 0) <= DataValidShift(DataValidShift'length-1 downto 1);
           
            --Coupling Addends to the first calculation stage of the adder tree
            for m in 0 to c_NumAddends-1 loop
               CalcStages(0,m).re(Addends(m).re'range) <= Addends(m).re;
               CalcStages(0,m).re(CalcStages(0,m).re'length-1 downto Addends(m).re'length) <= (others=> Addends(m).re(Addends(m).re'length-1));
            
               CalcStages(0,m).im(Addends(m).im'range) <= Addends(m).im;
               CalcStages(0,m).im(CalcStages(0,m).im'length-1 downto Addends(m).im'length) <= (others=> Addends(m).im(Addends(m).im'length-1));
            end loop;
            
            --at first stage the number of addends left is the same as number of addends
            v_n := c_NumAddends;
            --rest of the adder tree
            for i in 1 to logCeil(2,c_NumAddends) loop
               v_nNext := 0;
               for j in 0 to v_n/2-1 loop
                  CalcStages(i,j).re <= CalcStages(i-1,j).re + CalcStages(i-1,(v_n/2)*2-j-1).re;
                  CalcStages(i,j).im <= CalcStages(i-1,j).im + CalcStages(i-1,(v_n/2)*2-j-1).im;
                  --for every addition done in one level, there is an addend in the next level
                  v_nNext := v_nNext+1;
               end loop;
               --if there is an odd number of addends, one is transfered directly to the next stage
                  --and there is one extra addend at next stage
               if (v_n/2 /= (v_n+1)/2) then
                  CalcStages(i,(v_n+1)/2-1) <= CalcStages(i-1,v_n-1);
                  v_nNext := v_nNext+1;
               end if;
               v_n := v_nNext;
            end loop;
           
         
         end if; 
      end if;
   end process; 
   
   ----------------------------------------------------------------------------
   -- Instantiations
   ----------------------------------------------------------------------------

end rtl;

