--File: ComplexMatrixVecMul1610_rtl.vhd (vhdl)
--Hermitian(Vec)*Matrix*Vec operation
--(c) Kongsberg Defence & Aerospace 2015
--$Id: ComplexHVecMatrixVecMul1610_rtl.vhd 14777 2015-03-24 10:58:36Z mathiab $
--Author: Mathias By
--
--Assumptions: 
-- <Add Assumptions>
--Description: 
-- The ComplexHVecMatrixVecMul1610 module takes a 16 bit complex matrix and a 10 bit complex vector and
-- calculates the matrix product of the hermitian conjugate of the vector with the matrix and then takes
-- the vector multiplication between the result of the matrix multiplication and the input vector to obtain
-- the result. The input can be changed every 10 clk cycles and a ResultValid signal reflects the state that the 
-- DataValid signal had when a finished calculation started. To save logic both a slow and a fast muliplier is used.
-- The fast multiplier has a latency dependent on the length of the vectors and is fully pipelined. The slow
-- mulitplier has a delay equal to the length of the vectors. One fast muliplier is used to make one
-- vector element available at each cycle from the H(vector)*matrix operation, the slow multiplier handles this one element
-- per cycle to perform the ResVector*Vector operation.


architecture rtl of ComplexHVecMatrixVecMul1610 is

   ----------------------------------------------------------------------------
   -- Types, constants and signals
   ----------------------------------------------------------------------------
      --slow multiplier
      signal Vector16S : t_1xAT16; --16 bit vector input (Result vector)
      signal Vector10S : t_1xAT10; --10 bit vector input    
      signal DataStartS   : std_logic; --Start slow multiplier
      signal ResultValidS : std_logic;
      
      --fast multiplier
      signal Vector16F : t_1xAT16; --16 bit vector input (from matrix)
      signal Vector10F : t_1xAT10; --10 bit vector input
      signal ResultF   : t_Complex20;
      signal DataValidF   : std_logic; -- will be reflected in ResultValidF for same data set
      signal ResultValidF : std_logic;
      
      --Counters to keep track of which element that is handled and that is ready
      signal InputElementNumberF : integer range 0 to c_AT;
      signal OutputElementNumberF : integer range 0 to c_AT;

      --todo TESTING
      --signal ShiftReg : t_ATxAT16; 
      


   ----------------------------------------------------------------------------
   -- Procedures and functions
   ----------------------------------------------------------------------------

begin

   ----------------------------------------------------------------------------
   -- Combinatorial processes
   ----------------------------------------------------------------------------
   --When result is ready from the slow multiplier, then the result for the whole operation is ready
   ResultValid <= ResultValidS;
      
    
   --hermitian Conjugate of vector for the hermitian(vector)*Matrix multiplication
   copling : for i in 0 to c_At-1 generate
      Vector10F(i).re <= Vector10S(i).re;
      Vector10F(i).im <= -Vector10S(i).im;
   end generate;

   ----------------------------------------------------------------------------
   -- Clocked processes
   ----------------------------------------------------------------------------
   
   p_MyProcess : process(Clk) is
   
   begin
      if rising_edge(Clk) then
         DataValidF <= '0';
         DataStartS <= '0';
         if Rst_n = '0' then
            Ready <= '1';
            InputElementNumberF <= 0;
            OutputElementNumberF <= 0;
            Vector10S <= (others => (others => (others => '0') ) );
            Vector16S <= (others => (others => (others => '0') ) ); 
            Vector16F <= (others => (others => (others => '0') ) ); 
         else
            --fast multiplier
            Vector16F <= InputMatrix(InputElementNumberF);
            if InputElementNumberF = 0 then
               --start operation if steady state and data is valid
               if DataValid = '1' then
                  InputElementNumberF <= 1;
                  Ready <= '0';
                  DataValidF <= '1';
                  --When data is valid the input vector of the slow multiplier is stored to be the input vector.
                  Vector10S <= InputVector;
               end if;
            else
               -- when last element is put into fast multiplier the module is ready for new input
               if InputElementNumberF = c_AT -1 then
                  Ready <= '1';
                  InputElementNumberF <= 0;
               else
                  InputElementNumberF <= InputElementNumberF + 1;
               end if;
            end if;
            
            --todo TESTING
            ----fast multiplier
            --Vector16F <= ShiftReg(0);
            --ShiftReg <= InputMatrix; 
            --if InputElementNumberF = 0 then
            --   --start operation if steady state and data is valid
            --   if DataValid = '1' then
            --      InputElementNumberF <= 1;
            --      Ready <= '0';
            --      DataValidF <= '1';
            --      --When data is valid the input vector of the slow multiplier is stored to be the input vector.
            --      Vector10S <= InputVector;
            --      for i in 0 to c_AT -2 loop
            --         ShiftReg(i) <= ShiftReg(i+1);
            --      end loop;
            --   end if;
            --else
            --   for i in 0 to c_AT -2 loop
            --      ShiftReg(i) <= ShiftReg(i+1);
            --   end loop;
            --   -- when last element is put into fast multiplier the module is ready for new input
            --   if InputElementNumberF = c_AT -1 then
            --      Ready <= '1';
            --      InputElementNumberF <= 0;
            --   else
            --      InputElementNumberF <= InputElementNumberF + 1;
            --   end if;
            --end if;
            --todo TESTING end


            --slow mulitplier
            if OutputElementNumberF = 0 then
               --start operation if wait state and result is valid from fast multiplier
               if ResultValidF = '1' then
                  OutputElementNumberF <= 1;
                  DataStartS <= '1';
                  --the result of the 16*10 bit H(Vec)*Matr operation is truncated to 16 bits
                  Vector16S(OutputElementNumberF).re <= ResultF.re(ResultF.re'length-1 downto ResultF.re'length-Vector16S(0).re'length);         
                  Vector16S(OutputElementNumberF).im <= ResultF.im(ResultF.re'length-1 downto ResultF.re'length-Vector16S(0).re'length);         
               end if;
            --as long as not first or last element
            elsif OutputElementNumberF /= c_AT-1 then
               OutputElementNumberF <= OutputElementNumberF + 1;
               Vector16S(OutputElementNumberF).re <= ResultF.re(ResultF.re'length-1 downto ResultF.re'length-Vector16S(0).re'length);         
               Vector16S(OutputElementNumberF).im <= ResultF.im(ResultF.re'length-1 downto ResultF.re'length-Vector16S(0).re'length);         
            --last element, go back to wait state
            else
               OutputElementNumberF <= 0;
               Vector16S(OutputElementNumberF).re <= ResultF.re(ResultF.re'length-1 downto ResultF.re'length-Vector16S(0).re'length);         
               Vector16S(OutputElementNumberF).im <= ResultF.im(ResultF.re'length-1 downto ResultF.re'length-Vector16S(0).re'length);         
            end if;
         end if; 
      end if;
   end process; 
   
   ----------------------------------------------------------------------------
   -- Instantiations
   ----------------------------------------------------------------------------
   
   --slow multiplier   
      i_ComplexVectorMul1610Slow : entity GENETICALGLIB.ComplexVectorMul1610Slow(rtl)
         port map (
         Rst_n => Rst_n,
         Clk   => Clk,
         Vector16 => Vector16S,
         Vector10 => Vector10S,
         Result   => Result,
         DataStart   => DataStartS,
         ResultValid => ResultValidS
      );
   
   --fast multiplier
      i_ComplexVectorMul1610 : entity GENETICALGLIB.ComplexVectorMul1610(str)
         port map (
         Rst_n => Rst_n,
         Clk   => Clk,
         Hermit16 => Vector16F,
         Vector10 => Vector10F,
         Result   => ResultF,
         DataValid   => DataValidF,
         ResultValid => ResultValidF
      );

end rtl;

