--File: ComplexVectorMul1610Slow_rtl.vhd (vhdl)
--Slow 16 bit by 10 bit complex vector multiplier 
--(c) Kongsberg Defence & Aerospace 2015
--$Id: ComplexVectorMul1610Slow_rtl.vhd 14777 2015-03-24 10:58:36Z mathiab $
--Author: Mathias By
--
--Assumptions: 
-- <Add Assumptions>
--Description: 
-- The ComplexVectorMul1610Slow module perform a vector multiplication between the 16 bit signed complex vector Vector16
-- and the 10 bit signed complex vector Vector16. The vector multiplication is slow, the module uses one complex multiplier and
-- two accumulators. One complex multiplier (throughput) is done per cycle and the accumulators add up the result. This means that
-- the vector multiplication delay is proportional to the number of elements in the vectors. The resource usage is low the application.


architecture rtl of ComplexVectorMul1610Slow is

   COMPONENT Accumulator
   PORT (
      b : IN STD_LOGIC_VECTOR(15 DOWNTO 0);
      clk : IN STD_LOGIC;
      sclr : IN STD_LOGIC;
      q : OUT STD_LOGIC_VECTOR(19 DOWNTO 0)
   );
   END COMPONENT;
--------------------------------------------------------------------------
   -- Types, constants and signals
   ----------------------------------------------------------------------------
   signal AccClear      : std_logic ; --used to clear the accumulators
   --Accumulator inputs
   signal RealAccInput  : signed(Vector16(1).re'length-1 downto 0);
   signal ImagAccInput  : signed(Vector16(1).im'length-1 downto 0);
   --Multiplier operands
   signal MultiplicandA : t_Complex16; 
   signal MultiplicandB : t_Complex10;
   
   signal ElementNumber : integer range 0 to Vector16'length-1; --used to keep track of which multiplicands to push to multiplier
   Signal ResultValidReg: std_logic_vector (Vector16'length + c_MulLatency downto 0); --used to generate ResultValid and AccClear

   ----------------------------------------------------------------------------
   -- Procedures and functions
   ----------------------------------------------------------------------------

begin

   ----------------------------------------------------------------------------
   -- Combinatorial processes
   ----------------------------------------------------------------------------
   ResultValid <= ResultValidReg(ResultValidReg'length-1);
   --The Accumulators is cleared one clock cycle before the first result is ready from the multiplier
   AccClear <= ResultValidReg(c_MulLatency-1);

   ----------------------------------------------------------------------------
   -- Clocked processes
   ----------------------------------------------------------------------------
   
   p_ModuleSynchronisation : process(Clk) is 
   begin
      if rising_edge(Clk) then
         if Rst_n = '0' then
            ElementNumber <= 0;
            MultiplicandB <= (others => (others => '0')); 
            MultiplicandA <= (others => (others => '0')); 
            ResultValidReg   <= (others => '0');
         else
            --Assigning The  multiplicants
            MultiplicandB <= Vector10(ElementNumber); 
            MultiplicandA <= Vector16(ElementNumber); 
            --if it is no new data set, then shift the ResultValidReg and fill with a zero bit
            ResultValidReg(0) <= '0';
            ResultValidReg(ResultValidReg'length-1 downto 1) <= ResultValidReg(ResultValidReg'length-2 downto 0);
            
            --Control of the multiplications
            if ElementNumber = 0 then
               --if in wait state and new data set arrive, then fill ResultValidReg with a '1' bit and start multiplication
               if DataStart = '1' then
                  ElementNumber <= 1;
                  ResultValidReg(0) <= '1';
               end if;
            --Multiplications finished
            elsif ElementNumber = Vector16'length-1 then
               ElementNumber <= 0;
            else
               ElementNumber <= ElementNumber + 1;
            end if;
         end if; 
      end if;
   end process; 
   
   ----------------------------------------------------------------------------
   -- Instantiations
   ----------------------------------------------------------------------------
   --Accumulator for the real part of the result
   i_RealPartAcc : Accumulator
      PORT MAP (
      b => std_logic_vector(RealAccInput),
      clk => Clk,
      sclr => AccClear,
      signed(q) => Result.re
   );
   
   --Accumulator for the imaginary part of the result
   i_ImagPartAcc : Accumulator
      PORT MAP (
      b => std_logic_vector(ImagAccInput),
      clk => Clk,
      sclr => AccClear,
      signed(q) => Result.im
   );
   
   -- Complex multiplier
   i_ComplexMul16b10b : entity GENETICALGLIB.ComplexMul16b10b(rtl)
      port map (
      Rst_n => Rst_n,
      Clk   => Clk,
      A  => MultiplicandA,
      B     => MultiplicandB,
      Res.re   => RealAccInput,
      Res.im   => ImagAccInput
   );
end rtl;

