--File: CostFunction_rtl.vhd (vhdl)
-- Calculate H(W)*M*W - H(W)*C - H(C)*W + P1, Latency 18 Delay 10.
--(c) Kongsberg Defence & Aerospace 2015
--$Id: CostFunction_rtl.vhd 14777 2015-03-24 10:58:36Z mathiab $
--Author: Mathias By
--
--Assumptions: 
-- <Add Assumptions>
--Description: 
-- The cost function module calculates the expression Hermitian(W)*InputMatrix*W -
-- Hermitian(W)*C - Hermitian(C)*W + P1 where InputMatrix is a complex signed matrix,
-- W or InputVectorW is a 10 bit signed complex vector and C or InputVectorC is a 
-- 16 bit signed complex vector. P1 is a 20 bit signed complex number. the result is
-- truncated into 20 bit. The module delay is dependant on vector and matrix dimension
-- , the module can handle new input every 10 clock cycles. The module has a DataValid,
-- Ready and a ResultValid signal to avoid corrupt data.


architecture rtl of CostFunction is

   ----------------------------------------------------------------------------
   -- Types, constants and signals
   ----------------------------------------------------------------------------
   signal DelayedHermitianVectorW  : t_1xAT10;
   signal DelayedVectorW : t_1xAT10;
   
   signal DelayedHermitianVectorC  : t_1xAT16;
   signal DelayedVectorC : t_1xAT16;

   signal DelayedStartSlowMul : std_logic;

   signal partialResult1 : t_Complex20;
   signal partialResult2 : t_Complex20;

   
   --HVecMatrixVec signals
   signal MatrixDataValid   : std_logic;
   signal HVMatrixV         : t_Complex20;
   signal MatrixResultValid : std_logic;
   signal MatrixMulReady    : std_logic;

   --Slow Vector Mul signals
   signal H1610   : t_Complex20;
   signal N16H10   : t_Complex20;
   signal DataStartMultipliers : std_logic;
   signal ResultValidFromVectorMultipliers  : std_logic;


   signal DelayedValid : std_logic;
   ----------------------------------------------------------------------------
   -- Procedures and functions
   ----------------------------------------------------------------------------

begin
   
   ----------------------------------------------------------------------------
   -- Combinatorial processes
   ----------------------------------------------------------------------------
   

   ----------------------------------------------------------------------------
   -- Clocked processes
   ----------------------------------------------------------------------------
   
   p_MyProcess : process(Clk) is
   
   begin
      if rising_edge(Clk) then
         if Rst_n = '0' then
            DelayedVectorW   <= (others => (others => (others => '0'))); 
            DelayedVectorC   <= (others => (others => (others => '0'))); 
            DelayedStartSlowMul <= '0';
            
            partialResult1 <= (others => (others => '0') ); 
            partialResult2 <= (others => (others => '0') );
             
            DelayedValid <= '0';
            ResultValid <= '0'; 
         else
            --delaying MatrixResultValid and use it as Result valid, after the matrix
               --result is valid there is two clock cycle of calculations before
               --the result is valid. 
            DelayedValid <= MatrixResultValid;
            ResultValid  <= DelayedValid;
            
            --Calculating H(W)*C + H(C)*W When these calculations are ready from the multipliers
            if ResultValidFromVectorMultipliers = '1' then
               partialResult1.re <= to_signed(to_integer(H1610.re(c_NumBitsSum-1 downto 4)),c_NumBitsSum)+to_signed(to_integer(N16H10.re(c_NumBitsSum-1 downto 4)),c_NumBitsSum); 
               partialResult1.im <= to_signed(to_integer(H1610.im(c_NumBitsSum-1 downto 4)),c_NumBitsSum)+to_signed(to_integer(N16H10.im(c_NumBitsSum-1 downto 4)),c_NumBitsSum);
            end if;
            
            --Calculates P1 + H(W)*IM*W
            partialResult2.re <= to_signed(to_integer(P1.re(c_NumBitsSum-1 downto 4)),c_NumBitsSum)+HVMatrixV.re;
            partialResult2.im <= to_signed(to_integer(P1.im(c_NumBitsSum-1 downto 4)),c_NumBitsSum)+HVMatrixV.im;
            
            --Subtract the precalculated partial results to form (P1 + H(W)*IM*W) - (H(W)*C + H(C)*W) 
            Result.re <= partialResult2.re - partialResult1.re; 
            Result.im <= partialResult2.im - partialResult1.im;
            
            --preparing signals for the vector-vector muliplications, all signals are delayed one clock/ uses one clock on operation
            for i in 0 to c_AT - 1 loop 
               DelayedHermitianVectorW(i).re <=  InputVectorW(i).re; 
               DelayedHermitianVectorW(i).im <= -InputVectorW(i).im;
            end loop; 
            for i in 0 to c_AT - 1 loop 
               DelayedHermitianVectorC(i).re <=  InputVectorC(i).re; 
               DelayedHermitianVectorC(i).im <= -InputVectorC(i).im;
            end loop;
            
            DelayedVectorW <= InputVectorW;
            DelayedVectorC <= InputVectorC;
            
            DelayedStartSlowMul <= DataValid;
           
         end if; 
      end if;
   end process; 
   
   ----------------------------------------------------------------------------
   -- Instantiations
   ----------------------------------------------------------------------------
   --H(W)*IM*W
   i_ComplexHVecMatrixVecMul1610 : entity GENETICALGLIB.ComplexHVecMatrixVecMul1610(rtl)
      port map (
      Rst_n       => Rst_n,
      Clk         => Clk,
      InputMatrix => InputMatrix,
      InputVector => InputVectorW,
      DataValid   => DataValid,
      Result      => HVMatrixV,
      ResultValid => MatrixResultValid,
      Ready       => Ready
   );
   --H(C)*W
   i_ComplexVectorMul1610Slow_H1610 : entity GENETICALGLIB.ComplexVectorMul1610Slow(rtl)
      port map (
      Rst_n => Rst_n,
      Clk   => Clk,
      Vector16 => DelayedHermitianVectorC,
      Vector10 => DelayedVectorW,
      Result   => H1610,
      DataStart   => DelayedStartSlowMul,
      ResultValid => ResultValidFromVectorMultipliers
   );
   --H(W)*C
   i_ComplexVectorMul1610Slow_N16H10 : entity GENETICALGLIB.ComplexVectorMul1610Slow(rtl)
      port map (
      Rst_n => Rst_n,
      Clk   => Clk,
      Vector16 => DelayedVectorC,
      Vector10 => DelayedHermitianVectorW,
      Result   => N16H10,
      DataStart   => DelayedStartSlowMul,
      ResultValid => open --since the vector multipliers is started at the same
   );                         --time there is only need for one ResultValid signal.


end rtl;

