--File: DelayDataBuffer_rtl.vhd (vhdl)
--specialised buffer that controls the data flow of the genetic algorithm
--(c) Kongsberg Defence & Aerospace 2015
--$Id: DelayDataBuffer_rtl.vhd 14777 2015-03-24 10:58:36Z mathiab $
--Author: Mathias By
--
--Description: 
--DelayDataBuffer is a buffer that is filled with individuals when the dataValid signal is high. 
--It transmits data as two individuals every 4th clock cycle and set NewPopOut high every c_PopSize clock cycles   


architecture rtl of DelayDataBuffer is

   ----------------------------------------------------------------------------
   -- Types, constants and signals
   ----------------------------------------------------------------------------
   signal Migrants      : t_1xAT10Array; --Used for migration between the populations
   signal DataOutReg    : t_1xAT10Array; -- Output register
   signal CounterIn     : integer range 0 to 63; --used to circulate memory usage
   signal CounterOut    : integer range 0 to c_PopSize-1; --used to keep track of the transmission of individuals
   signal OutState      : integer range 0 to c_NumOfPop-1; --Also used to keep track of transmission
   signal LastDataValid : std_logic;   -- Used for edge detection

   --block memory signals
   signal WriteEnable : std_logic;
   signal WriteAddr   : std_logic_vector(5 downto 0) ;
   signal WriteData   : std_logic_vector(179 downto 0);
   signal ReadAddr    : std_logic_vector(5 downto 0);
   signal ReadData    : std_logic_vector(179 downto 0);

   


   ----------------------------------------------------------------------------
   -- Procedures and functions
   ----------------------------------------------------------------------------

begin

   ----------------------------------------------------------------------------
   -- Combinatorial processes
   ----------------------------------------------------------------------------
   --Circulating block memory write address
   WriteAddr <= std_logic_vector(to_unsigned(CounterIn,6));
   
   --maping the bits of the weigts to the memory's write data.
   Couplings : for i in 0 to c_AT-1 generate
      Couplings2: for j in 0 to c_NumBitsLow-1  generate
         WriteData(c_NumBitsLow*i*2+j)       <= DataIn(i).re(j);
         WriteData(c_NumBitsLow*(i*2+1) + j) <= DataIn(i).im(j);
      end generate;
   end generate;
   
   WriteEnable <= DataInValid;
   

   ----------------------------------------------------------------------------
   -- Clocked processes
   ----------------------------------------------------------------------------
   
   p_MyProcess : process(Clk) is
      variable v_CounterOutVector : std_logic_vector(9 downto 0);
   begin
      if rising_edge(Clk) then
         NewPopOut <= '0'; 
         --Vector version of CounterOut
         v_CounterOutVector := std_logic_vector(to_unsigned(CounterOut,10));
         if Rst_n = '0' then
            OutState <= 0;
            CounterOut <= c_PopSize - 1;
            CounterIn <= (c_NumOfPop-1)*c_PopSize/2; --becuase the 
            Migrants <= (others => (others => (others => (others => '0'))));
            DataOutReg <= (others => (others => (others => (others => '0')))); 
            LastDataValid <= '0';
            CurrWeights <= (others => (others => (others => '0')));
         else
            if DataInValid = '1' then
               --Write current valid data and increment address (CounterIn) 
               CounterIn <= CounterIn + 1;
               
               --Output for GeneticAlg is best individual of the first population
               if CounterIn = c_PopSize/2-1 then
                  CurrWeights  <= DataIn;
               end if;
               --The best two individuals from each population migrates to the next population through the Migrants signal
               for i in 1 to c_NumOfPop loop
                  if CounterIn = c_PopSize/2*i - 2 then
                     Migrants(0) <= DataIn;
                  elsif CounterIn = c_PopSize/2*i - 1 then
                     Migrants(1) <= DataIn;
                  end if;
               end loop;
            else
               --End of last population
               if CounterIn > (c_NumOfPop-1)*c_PopSize/2 then
                  CounterIn <= 0;
               end if;
            end if;
            --Outstate keeps hold of which population that is transmitted 
            if CounterOut = 0 then
               if OutState = c_NumOfPop - 1 then
                  OutState <= 0;
               else
                  OutState <= OutState +1;
               end if;
               CounterOut <= c_PopSize-1;
            else
               --CounterOut counts down from c_PopSize -1 downto 0
               CounterOut <= CounterOut -1;
            end if;

            -- For the output, the data have to be read from the block memory. The output format is two individuals
               --every 4th clock cycle. The NewPopOut signal is set if the two first individuals of a population
               --is set to the output  
            if v_CounterOutVector(1 downto 0) = "11" then
               --Send read request for individual 1 to block memory
               ReadAddr <= std_logic_vector(to_unsigned(Outstate*c_PopSize/2 + (CounterOut/4)*2+1,6));
               --Output the data read from the block memory during the last four clock cycles
               DataOut <= DataOutReg;
               -- last set of individuals from the previous population is the migrants
               if CounterOut = c_PopSize - 1 then
                  DataOut <= Migrants;
               elsif CounterOut = c_PopSize - 5 then
                  --When the first set of current population is sent the NewPopOut signal is set high for one clock cycle
                  NewPopOut <= '1'; 
               end if;
            elsif v_CounterOutVector(1 downto 0) = "10" then
               --Send read request for individual 2 to block memory
               ReadAddr <= std_logic_vector(to_unsigned(Outstate*c_PopSize/2 + (CounterOut/4)*2,6));
            elsif v_CounterOutVector(1 downto 0) = "01" then
               --Receiving individual 1 from block memory 
               for i in 0 to c_AT-1 loop --translating from memory ReadData back to weights
                  for j in 0 to c_NumBitsLow-1  loop
                     DataOutReg(0)(i).re(j) <= ReadData(c_NumBitsLow*i*2+j);       
                     DataOutReg(0)(i).im(j) <= ReadData(c_NumBitsLow*(i*2+1) + j);
                  end loop;
               end loop;
            else
               --Receiving individual 2 from block memory 
               for i in 0 to c_AT-1 loop --translating from memory ReadData back to weights
                  for j in 0 to c_NumBitsLow-1  loop
                     DataOutReg(1)(i).re(j) <= ReadData(c_NumBitsLow*i*2+j);       
                     DataOutReg(1)(i).im(j) <= ReadData(c_NumBitsLow*(i*2+1) + j);
                  end loop;
               end loop;
            end if;
         end if; 
      end if;
   end process; 
   

   ----------------------------------------------------------------------------
   -- Instantiations
   ----------------------------------------------------------------------------
   i_BlockMemory : entity GENETICALGLIB.Blockmemory180x64(Blockmemory180x64_a)
      PORT MAP (
         clka =>  Clk,
         wea(0) => WriteEnable,
         addra => WriteAddr,
         dina =>  WriteData,
         clkb =>  Clk,
         addrb => ReadAddr,
         doutb => ReadData);

end rtl;

