--File: SerialCostFunction_rtl.vhd (vhdl)
--The SerialCostFunction is a fully paralellised cost function with latency 19
--(c) Kongsberg Defence & Aerospace 2015
--$Id: SerialCostFunction_rtl.vhd 14777 2015-03-24 10:58:36Z mathiab $
--Author: Mathias By
--
--Description: 
--The SerialCostFunction module uses 10 Cost function cores which have a latency of 18 and a
-- a delay of 10. One core get a new input every clock cycle and a result is therefore ready
-- every clock cycle.


architecture rtl of SerialCostFunction is

   ----------------------------------------------------------------------------
   -- Types, constants and signals
   ----------------------------------------------------------------------------
   signal InputVectorWArray   : t_1xAT10ArrayUnCon(0 to c_AT); --signal to hold input weights
   signal DelayInputVectorWArray   : t_1xAT10ArrayUnCon(0 to c_AT); --signal to delay input weights to meet timing

   constant c_CostFLatency : integer := 18; --latency of cost function
   
   type t_AOfSlv is array(natural range <>) of std_logic_vector(6 downto 0); 
   signal DelayWAddressBuffer : t_AOfSlv(0 to c_CostFLatency); --to keep track of which cost belongs to which weights
   
   signal DataValidVector     : std_logic_vector(0 to c_AT); --signal used to give data valid to the different cost function cores
   signal DelayDataValidVector     : std_logic_vector(0 to c_AT); --used to delay data valid vector to meet timing
   
   type t_AOfComplex20 is array(0 to c_AT) of t_Complex20;
   signal ResultArray         : t_AOfComplex20; --signal to hold the results from the different cost function cores
   
   signal ResultValidVector   : std_logic_vector(0 to c_AT); --used to keep the result valid signals from the different cost function cores
   signal NewPopOutReg        : std_logic_vector(c_CostFLatency downto 0); --used to time the NewPopOut signal to be sent at the same time as the
                                                                              --weights that was received with the NewPopIn signal
   signal Counter     : integer range 0 to 4*c_PopSize-1; --Counter used to circulate memory usage
   ----------------------------------------------------------------------------
   -- Procedures and functions
   ----------------------------------------------------------------------------

begin
   
   ----------------------------------------------------------------------------
   -- Combinatorial processes
   ----------------------------------------------------------------------------
      
   
   OutputWeightAddress <= std_logic_vector(to_unsigned(Counter,7));  --The address of the block memory that is used to store
                                                                        --the weight set are determined by Counter   
   OutputCostAddress <= DelayWAddressBuffer(c_CostFLatency); --The address of the weights which are used to find the current
                                                               --output cost comes from shift register
   NewPopOut <= NewPopOutReg(c_CostFLatency); --NewPopOut comes from a shiftregister which delays NewPopIn

   ----------------------------------------------------------------------------
   -- Clocked processes
   ----------------------------------------------------------------------------
   
   p_MyProcess : process(Clk) is
      variable v_Counter : integer range 0 to c_AT; --Counter used to circulate cost function core usage
   begin
      if rising_edge(Clk) then
         if Rst_n = '0' then
            Counter <= 0;      
            v_Counter := 0;
            DelayWAddressBuffer <= (others =>(others => '0'));
            NewPopOutReg <= (others => '0'); 
            DataValidVector <= (others => '0');
         else
            --Delay weight array and data valid vector to meet timing
            DelayInputVectorWArray <= InputVectorWArray;
            DelayDataValidVector <= DataValidVector;
            --Counter to keep circulate use of memory 
            if Counter = 4*c_PopSize-1 then
               Counter <= 0;
            else
               Counter <= Counter+1;
            end if;
            
            --Shifting registers which is used to sync NewPopOut and the address where the weights are in the block memory
               --with the output cost
            for i in 1 to c_CostFLatency  loop
               DelayWAddressBuffer(i) <= DelayWAddressBuffer(i-1);
               NewPopOutReg(i) <= NewPopOutReg(i-1);
            end loop;
            NewPopOutReg(0) <= NewPopIn;
            DelayWAddressBuffer(0) <= std_logic_vector(to_unsigned(Counter,7));
            
            --Setting data valid high for the cost function core which the input weights are pushed to
            DataValidVector <= (others => '0');
            DataValidVector(v_Counter) <= '1';
            
            --pushing the input weights to one of the cost function cores
            InputVectorWArray(v_Counter) <= InputVectorW;
            
            --Setting any valid result from the cost function cores to the output cost (Result)
            for i in 0 to c_AT loop
               if ResultValidVector(i) = '1' then
                  if i = 0 then
                     Result <= ResultArray(c_AT);
                  else   
                     Result <= ResultArray(i-1);
                  end if;
               end if;   
            end loop; 
            --Circulating the usage of the cost function cores
            if v_Counter = c_AT then
               v_Counter := 0;
            else
               v_Counter := v_Counter + 1;
            end if;
         end if; 
      end if;
   end process; 
   
   ----------------------------------------------------------------------------
   -- Instantiations
   ----------------------------------------------------------------------------
   --Cost function cores
   CostFunctionCores : for i in 0 to c_AT generate
      i_CostFunction : entity GENETICALGLIB.CostFunction(rtl)
         port map (
         Rst_n => Rst_n,
         Clk   => Clk,
         InputMatrix    => InputMatrix,
         InputVectorC   => InputVectorC,
         InputVectorW   => DelayInputVectorWArray(i),
         DataValid      => DelayDataValidVector(i),
         P1             => P1,
         Result         => ResultArray(i),
         ResultValid    => ResultValidVector(i),
         Ready          => open
      );
   end generate; 
end rtl;

