--File: SortingRegister_rtl.vhd (vhdl)
--Sorts one input set per cycle using insertion sort. Transmitts only c_PopSize input sets per sort.
--(c) Kongsberg Defence & Aerospace 2015
--$Id: SortingRegister_rtl.vhd 14777 2015-03-24 10:58:36Z mathiab $
--Author: Mathias By
--
--Description: 
--The SortingRegister module take an address and a cost in every clock cycle. This input is sorted by its cost
--using insertion sort. The module can handle one new input every clock cycle, but only c_PopSize/2 elements are kept during the
--sort. When the NewPopIn signal goes high the sorted addresses are sent out at the AddressOut output one per clock cycle and
-- with the kept individual with the highest cost first. The TransmittingResult signal is kept high while the transmitted reslut is
--valid.


architecture rtl of SortingRegister is

   ----------------------------------------------------------------------------
   -- Types, constants and signals
   ----------------------------------------------------------------------------

   type t_ShiftRegElement is record --element containing both cost and address
      cost    : signed(19 downto 0);
      address : std_logic_vector (6 downto 0);
   end record;
   type t_ShiftReg is array(c_PopSize/2-1 downto 0) of t_ShiftRegElement; --sorting shift register
   
   signal ShiftReg  : t_ShiftReg; --sorting shift register
   signal ForcePush : std_logic_vector(c_PopSize/2-1 downto 0); --shit register used for transmission
   signal LastCost      : signed(19 downto 0); --used to delay cost
   signal LastAddress   : std_logic_vector (6 downto 0); --used to delay address
   ----------------------------------------------------------------------------
   -- Procedures and functions
   ----------------------------------------------------------------------------

begin

   ----------------------------------------------------------------------------
   -- Combinatorial processes
   ----------------------------------------------------------------------------
   TransmittingResult <= ForcePush(0);

   ----------------------------------------------------------------------------
   -- Clocked processes
   ----------------------------------------------------------------------------
   
   p_MyProcess : process(Clk) is
   
   begin
      if rising_edge(Clk) then
         if Rst_n = '0' then
            ForcePush <= (others => '0'); 
            LastCost   <= (others => '0') ;            
            LastAddress <= (others => '0') ;
         else
            --delaying cost and address
            LastCost   <= InputCost ;            
            LastAddress <= AddressIn ;
            
            if NewPopIn = '1' then
               --when a new population arrives the transmission register is filled with ones
               ForcePush <= (others => '1');
            else
               --otherwise the transmission register is shifted
               ForcePush <= '0' & ForcePush(c_PopSize/2-1 downto 1);
            end if;
            --handling first element
            if (ShiftReg(c_PopSize/2-1).cost > LastCost) or ForcePush(c_PopSize/2-1) ='1' then
               ShiftReg(c_PopSize/2-1).cost <= LastCost;
               ShiftReg(c_PopSize/2-1).address <= LastAddress;
            end if; 
            --rest of the elements
            for i in c_PopSize/2-2 downto 0 loop
               if ForcePush(i) = '1' then --if current element is forced to be shifted
                  if ForcePush(i+1) = '1' or ShiftReg(i+1).cost > LastCost then --if the previous element is forced to be shifted
                                                                                   --or previous element have larger cost than the element beeing filled in
                     --the current element should be the last element
                     ShiftReg(i).cost    <= ShiftReg(i+1).cost ;
                     ShiftReg(i).address <= ShiftReg(i+1).address ;
                  else
                     --if not the current element should be the element that is beeing filled in 
                     ShiftReg(i).cost <= LastCost;
                     ShiftReg(i).address <= LastAddress;
                  end if;
               else --if the current element is not forced to be shifted
                  --it should be replaced only if it is has a larger cost than the element that is beeing filled in
                  if ShiftReg(i).cost > LastCost then
                     --if the previous element also is larger than the element beeing filled in 
                     if ShiftReg(i+1).cost > LastCost then
                        --the current element should be set to the previous
                        ShiftReg(i).cost    <= ShiftReg(i+1).cost ;
                        ShiftReg(i).address <= ShiftReg(i+1).address ;
                     else
                        --if not the element that is beeing filled in should be placed in the current element
                        ShiftReg(i).cost <= LastCost;
                        ShiftReg(i).address <= LastAddress;
                     end if;
                  end if;
               end if;
            end loop;
            --the last element is always sent, so the TransmittingResult signal has to be used to see when the result is valid
            AddressOut <= ShiftReg(0).address;
         end if; 
      end if;
   end process; 
   
   ----------------------------------------------------------------------------
   -- Instantiations
   ----------------------------------------------------------------------------

end rtl;

